import ij.IJ;
import ij.ImageListener;
import ij.ImagePlus;
import ij.WindowManager;
import ij.gui.ImageCanvas;
import ij.gui.Overlay;
import ij.gui.PolygonRoi;
import ij.gui.Roi;
import ij.io.OpenDialog;
import ij.io.SaveDialog;
import ij.plugin.frame.PlugInFrame;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;


import hw.celldetection.*;


/*
 *	CellDetection
 * 		This plugin just put automatic and manual detected Roi on overlay.
 * 		And it can export vertices of polygonal Roi as TSV file.(also import)
 * 		The recorded Rois are measured by ImageJ.(just macro)
 *  
 *	20160719	version 0.8
 *		PxɋEFRoioverlay֓o^plugin
 *		toRoimanagersƂmeasure\Ƃ
 *	20160808	version 1.0
 *		manualo^@\ǉ
 *		tsvt@Cւ̓ǂݏ@\ǉ
 *		measure{^ǉ
 *			->̂RoiManageerAmeasure\ɁB
 *
 *	20160816	version 1.1
 * 		ċAgȂASYɂstackoverflow
 *		roi̊gɋNsi炭ARoi̓ɂ_󋵁jɂ Roi𖳎ݒ
 *			->ȂƂ܂\łȂ݂ASsBBB̂߂܂łv1.1ƂB
 *
 *	20160817
 *		stackoverflowƂɂ傫TCYF悤ƂĎԂ肪B
 *			->CellDetectorɂdetectSize(|S̒_)1000傫ꍇnullƂ邱Ƃŉ
 *		NbN܂overlayւ̓o^NtЂƂi܂@\ǉ
 *
 *	20160823
 *		savesfitspline؂
 *			->saveɁAremoveSplinesĂBȂspline̓_ۑĂ܂B
 *			
 */



@SuppressWarnings("serial")
public class CellDetection_ extends PlugInFrame implements MouseListener, ImageListener, MouseWheelListener, WindowListener,ItemListener{

	ImagePlus imp;
	ImageCanvas ic;

	Overlay currentOverlay;
	Overlay ol[];//stacksize
	Overlay saveOl[];//for save
	Overlay backupOl[];//stacksize
	
	Color strokeColor;
	
	JLabel intensityLimitLabel;
	JTextField intensityLimit;

	JLabel countLimitLabel;
	JTextField countLimit;

	JLabel radiusLabel;
	JTextField radius;
	
	JLabel expansionLabel;
	JTextField expansion;

	JLabel roughnessLabel;
	JTextField roughness;
	
	JLabel strokeWidthLabel;
	JTextField strokeWidth;
	
	ButtonGroup radioGroup;
	JRadioButton radioAutoDetect;
	JRadioButton radioManual;
	JRadioButton radioEdit;

	
	JCheckBox checkOverlay;
	JCheckBox checkSpline;
	JCheckBox checkAutoForwardT;
	
	JLabel colorBoxLabel;
	JComboBox<String> colorBox;
	String[] colors = {"White","Black","Red","Green","Blue","Yellow","Cyan","Magenta"};
	//String[] colors = IJ.getLuts();

	JComboBox<String> methodBox;
	String[] methods = {"MyMethod"};

	JButton buttonLoadTSV;
	JButton buttonSaveTSV;
	
	
	String title;
	String dir;
	String file;
	String export_extension = "tsv";
	String import_extension = "tsv";
	
	JButton buttonMeasure;
	JLabel noContent;
	
	int clicked_x;
	int clicked_y;
	
	public CellDetection_() {
		super("CellDetection_");
		// TODO Auto-generated constructor stub
		
		imp = WindowManager.getCurrentImage();

		if (imp == null) {
			IJ.noImage();
			return;
		}
		
		
		
		ic = imp.getCanvas();
		this.setListener();
		
		showPanel();
		this.overlayInitialize();

	}
	
	
	public void setListener(){
		ic.addMouseListener(this);
		ic.addMouseWheelListener(this);
		imp.getWindow().addWindowListener(this);
		ImagePlus.addImageListener(this);
	}
	
	public void removeListener(){
		ic.removeMouseListener(this);
		ic.removeMouseWheelListener(this);
		imp.getWindow().removeWindowListener(this);
		ImagePlus.removeImageListener(this);
	}
	
	public void setOverlayArray(Overlay[] o){
		ol = o;
		imp.updateAndDraw();
	}
	
	public void overlayInitialize(){

		int stackSize = imp.getStackSize();
		

		ol = new Overlay[stackSize];
		for(int cs = 0; cs < stackSize; cs++){
			ol[cs] = new Overlay();
		}
		strokeColor = convertColor(colorBox.getSelectedItem().toString());
		currentOverlay = ol[imp.getCurrentSlice()-1];
		currentOverlay.drawLabels(true);
		currentOverlay.setStrokeColor(strokeColor);
		currentOverlay.setLabelColor(strokeColor);
		imp.setOverlay(currentOverlay);
		backupOl = ol.clone();
		saveOl = ol.clone();
	}
	
	
	public void showPanel(){
		
        FlowLayout gd_layout = new FlowLayout();
    	//gd_layout.setAlignOnBaseline(true);
    	gd_layout.setAlignment(FlowLayout.LEFT);
        
        JPanel gd_panel = new JPanel(gd_layout);
    	gd_panel.setPreferredSize(new Dimension(265, 360));
    	
    	JPanel mainPanel = new JPanel((new GridLayout(12,2)));
    	mainPanel.setAlignmentX(JPanel.BOTTOM_ALIGNMENT);
    	
    	intensityLimitLabel = new JLabel("Intensity Limit:");
    	intensityLimit = new JTextField(String.valueOf(imp.getProcessor().getMax()/4));

    	countLimitLabel = new JLabel("Count Limit:");    	
    	countLimit = new JTextField("1");
    	
    	radiusLabel = new JLabel("Radius");
    	radius = new JTextField("2");
    	
    	roughnessLabel = new JLabel("Roughness");
    	roughness = new JTextField("2");
    	
    	expansionLabel = new JLabel("Expansion");
    	expansion = new JTextField("0");
    	
    	
    	strokeWidthLabel = new JLabel("StrokeWidth");
    	strokeWidth = new JTextField("0");

    	radioGroup = new ButtonGroup();
    	radioAutoDetect = new JRadioButton("AutoDetect");
    	radioManual = new JRadioButton("Manual");
    	radioEdit = new JRadioButton("Edit");
    	radioAutoDetect.addItemListener(this);    	
    	radioManual.addItemListener(this);    	
    	radioEdit.addItemListener(this);
    	
    	radioGroup.add(radioAutoDetect);
    	radioGroup.add(radioManual);
    	radioGroup.add(radioEdit);
    	radioAutoDetect.setSelected(true);
    	
    	checkOverlay = new JCheckBox("Overlay");
    	checkOverlay.setSelected(true);

    	checkSpline = new JCheckBox("Spline");
    	checkSpline.setSelected(true);
    	
    	colorBoxLabel = new JLabel("StrokeColor");
    	colorBox = new JComboBox<String>(colors);
    	colorBox.addItemListener(this);
    	
    	methodBox = new JComboBox<String>(methods);

    	buttonMeasure = new JButton("Measure");
    	buttonMeasure.addMouseListener(this);
    	//noContent = new JLabel("");
    	checkAutoForwardT = new JCheckBox("AutoForward-T");
    	checkAutoForwardT.setSelected(true);
    	
    	buttonLoadTSV = new JButton("Load TSV");
    	buttonLoadTSV.addMouseListener(this);
    	buttonSaveTSV = new JButton("Save TSV");
    	buttonSaveTSV.addMouseListener(this);
    	
    	mainPanel.add(intensityLimitLabel);
    	mainPanel.add(intensityLimit);

    	mainPanel.add(countLimitLabel);
    	mainPanel.add(countLimit);
    	
    	mainPanel.add(radiusLabel);
    	mainPanel.add(radius);
    	
    	mainPanel.add(expansionLabel);
    	mainPanel.add(expansion);
    	
    	mainPanel.add(roughnessLabel);
    	mainPanel.add(roughness);
    	
    	mainPanel.add(strokeWidthLabel);
    	mainPanel.add(strokeWidth);

    	mainPanel.add(colorBoxLabel);
    	mainPanel.add(colorBox);
    	
    	mainPanel.add(radioAutoDetect);
    	mainPanel.add(methodBox);
    	
    	mainPanel.add(radioManual);
    	mainPanel.add(checkSpline);

    	mainPanel.add(radioEdit);
    	mainPanel.add(checkOverlay);

    	mainPanel.add(buttonLoadTSV);
    	mainPanel.add(checkAutoForwardT);
    	
    	mainPanel.add(buttonSaveTSV);
    	mainPanel.add(buttonMeasure);

    	
    	gd_panel.add(mainPanel);
    	
    	this.add(gd_panel);
		this.pack(); //TCŶindow
		
		Point imp_point = imp.getWindow().getLocation();
		int imp_window_width = imp.getWindow().getWidth();
		//int imp_window_height = imp.getWindow().getHeight();

		double set_x_point = imp_point.getX() + imp_window_width;
		double set_y_point = imp_point.getY();
		
		this.setLocation((int)set_x_point, (int)set_y_point);

		this.setVisible(true);//this̕\

    	
	}	

	
	public boolean showSaveDialog(){
    	title = imp.getTitle();
    	dir = imp.getOriginalFileInfo().directory;
    	
    	SaveDialog sd = new SaveDialog("Export tsv file", dir, title, "." + export_extension);
    	
    	if (sd.getDirectory() == null || sd.getFileName() == null) {
    		return false;
    	}
    	dir = sd.getDirectory(); //save dialog@ɑÎɕύXB
    	file = sd.getFileName(); //save dialog@ɑÎɕύXB
    	return true;
		
	}

	public boolean showLoadDialog(){
    	title = imp.getTitle();
    	dir = imp.getOriginalFileInfo().directory;
    	
    	OpenDialog od = new OpenDialog("Import tsv file", dir, title + import_extension);
    	
    	if (od.getDirectory() == null || od.getFileName() == null) {
    		return false;
    	}
    	dir = od.getDirectory(); //save dialog@ɑÎɕύXB
    	file = od.getFileName(); //save dialog@ɑÎɕύXB
    	return true;		
	}
	
	
	public boolean detectCell(){

		boolean b = false;
		int cs = imp.getCurrentSlice();
		
		double intLimit = Double.valueOf(intensityLimit.getText());
		int cntLimit = Integer.valueOf(countLimit.getText());
		int radiusNum = Integer.valueOf(radius.getText());
		int expansionNum = Integer.valueOf(expansion.getText());
		int roughnessNum = Integer.valueOf(roughness.getText());
		int strokeWidthNum = Integer.valueOf(strokeWidth.getText());
		Color strokeC = convertColor(colorBox.getSelectedItem().toString());
		
		CellDetector cd = new CellDetector();
		cd.setImage(imp);
		cd.setStartPosition(clicked_x, clicked_y);
		cd.setCountLimit(cntLimit);
		cd.setIntensityLimit(intLimit);
		cd.setRadius(radiusNum);
		cd.setExpansion(expansionNum);
		cd.setRoughness(roughnessNum);
		cd.setStrokeWidth(strokeWidthNum);
		cd.setStrokeColor(strokeC);
		
		if(checkSpline.isSelected()){
			cd.setFitSpline(true);
		}else{
			cd.setFitSpline(false);
		}
		
		if(cd.detectNegativePosition()){
			//cd.drawNegativeRegion();
			Roi roi = cd.detectRoi();
			if(roi != null){
				imp.setRoi(roi);
				
				if(checkOverlay.isSelected()){
					Overlay overlay = imp.getOverlay();//RoiManegerԂĂƂ΍
					if(overlay == null){

						
						currentOverlay = backupOl[cs-1].duplicate();
						currentOverlay.drawLabels(true);
						imp.setOverlay(currentOverlay);
						
						overlay = imp.getOverlay();
					}
					overlay.drawLabels(true);
					overlay.setStrokeColor(strokeC);
					overlay.add(roi);
					b = true;
	
				}
			}
		}
		return b;
	}
	
	public boolean addOverlay(){
		boolean b = false;
		Roi r = imp.getRoi();
		int roiTypeID = 0;
		if(r != null){
		  roiTypeID = r.getType();
		}
		if((r != null)&&(ic.getCursor().getType() != Cursor.HAND_CURSOR)){
			//System.out.println(r.getTypeAsString());
			if(r.getState() != 0){ //0 == contructing, 3 == nomal
				if((roiTypeID == 2)||(roiTypeID == 6)){
					PolygonRoi pr;
					pr = (PolygonRoi)r;
					if(checkSpline.isSelected()){
						pr.fitSpline();
					}
					currentOverlay.add(pr);
				}else{
					currentOverlay.add(r);
				}
				currentOverlay.setStrokeColor(strokeColor);
				b = true;
			}
		}
		return b;
	}
	
	public Color convertColor(String s){
		Color c = Color.WHITE;
		
		if(s == "White"){
			c = Color.WHITE;
		}else if(s == "Black"){
			c = Color.BLACK;
		}else if(s == "Red"){
			c = Color.RED;
		}else if(s == "Green"){
			c = Color.GREEN;
		}else if(s == "Blue"){
			c = Color.BLUE;
		}else if(s == "Yellow"){
			c = Color.YELLOW;
		}else if(s == "Cyan"){
			c = Color.CYAN;
		}else if(s == "Magenta"){
			c = Color.MAGENTA;
		}

		return c;
	}
	
	public void forwardT(){
		if(checkAutoForwardT.isSelected()){
			int ct = imp.getT();
			imp.setT(ct + 1);
			imp.killRoi();
		}
	}
	
	
	@Override
	public void mouseClicked(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mousePressed(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseReleased(MouseEvent e) {

		if(e.getSource() == ic){
			if(radioAutoDetect.isSelected()){
				if((ic.getCursor().getType() != Cursor.HAND_CURSOR)){
					if(e.getSource() == ic){
						clicked_x = ic.offScreenX(e.getX());
						clicked_y = ic.offScreenY(e.getY());
						if(detectCell()){
							forwardT();
						}
					}
				}
			}else if(radioManual.isSelected()){
				if(addOverlay()){
					forwardT();
				}
			}
		
		}else if(e.getSource() == buttonSaveTSV){
				
			if(showSaveDialog()){
				
				String fileName = file;
				RoiTsvIO rtio = new RoiTsvIO(imp, dir, fileName);
				Overlay[] saveOl = new Overlay[ol.length];//splineݒ肪؂ꂿႤ
				for(int i = 0; i < ol.length; i++){
					saveOl[i] = ol[i].duplicate();
				}
				rtio.saveOverlayTSV(saveOl);
				
			}else{
				return;
			}

		}else if(e.getSource() == buttonLoadTSV){
			if(showLoadDialog()){
				String fileName = file;
				RoiTsvIO rtio = new RoiTsvIO(imp, dir, fileName);		
				setOverlayArray(rtio.loadOverlayTSV());
			}else{
				return;
			}
		}else if(e.getSource() == buttonMeasure){
			RoiManagerConverter rmc = new RoiManagerConverter(imp,ol);
			rmc.directMeasure();
		}else{
			
		}
		
	}

	@Override
	public void mouseEntered(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseExited(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void imageOpened(ImagePlus imp) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void imageClosed(ImagePlus imp) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void imageUpdated(ImagePlus imp) {

		int cs = imp.getCurrentSlice();

		currentOverlay = ol[cs -1];
		currentOverlay.drawLabels(true);
		imp.setOverlay(currentOverlay);
		
	}

	@Override
	public void mouseWheelMoved(MouseWheelEvent e) {
		imp.killRoi();
		if(e.getSource() == ic){
			int count = e.getWheelRotation();
			imp.setT(imp.getT() + count);
		}
	}

	@Override
	public void windowClosing(WindowEvent e) {
		removeListener();
		imp.setOverlay(null);
		this.close();
	}

	@Override
	public void windowOpened(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowClosed(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowActivated(WindowEvent e){
		if(this.isVisible()){
			changeSplineCheck();
		}
	}


	@Override
	public void itemStateChanged(ItemEvent e) {
		//System.out.println(e.getSource());
		if(e.getSource() == colorBox){
			strokeColor = convertColor(colorBox.getSelectedItem().toString());
			currentOverlay.setStrokeColor(strokeColor);	
			currentOverlay.setLabelColor(strokeColor);
			imp.updateAndDraw();
		}else if((e.getSource() == radioAutoDetect)||(e.getSource() == radioManual)||(e.getSource() == radioEdit)){
			if(this.isVisible()){
				changeSplineCheck();
			}
		}
	}
	
	public void changeSplineCheck(){
		
		if((radioManual.isSelected())||(radioEdit.isSelected())){
			String toolName = IJ.getToolName();
			if((toolName.equals("polygon"))||(toolName.equals("polyline"))){
				checkSpline.setEnabled(true);
			}else{
				checkSpline.setEnabled(false);
			}
		}else if(radioAutoDetect.isSelected()){
			checkSpline.setEnabled(true);
		}
	}
	
}